@@ -1,4 +1,4 @@
-/*  $Id: matrix.c,v 1.8 2003/04/07 16:27:10 ukai Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * matrix.h, matrix.c: Liner equation solver using LU decomposition.
  *
@@ -34,18 +34,12 @@
 
 #include "config.h"
 #include "matrix.h"
-#include <gc.h>
+#include "alloc.h"
 
 /* 
  * Macros from "fm.h".
  */
 
-#define New(type)       ((type*)GC_MALLOC(sizeof(type)))
-#define NewAtom(type)   ((type*)GC_MALLOC_ATOMIC(sizeof(type)))
-#define New_N(type,n)   ((type*)GC_MALLOC((n)*sizeof(type)))
-#define NewAtom_N(type,n)       ((type*)GC_MALLOC_ATOMIC((n)*sizeof(type)))
-#define Renew_N(type,ptr,n)   ((type*)GC_REALLOC((ptr),(n)*sizeof(type)))
-
 #define SWAPD(a,b) { double tmp = a; a = b; b = tmp; }
 #define SWAPI(a,b) { int tmp = a; a = b; b = tmp; }
 
@@ -125,7 +119,7 @@ LUfactor(Matrix A, int *indexarray)
  */
 
 int
-LUsolve(Matrix A, int *indexarray, Vector b, Vector x)
+LUsolve(Matrix A, const int *indexarray, Vector b, Vector x)
 {
     int i, dim = A->dim;
 
@@ -137,20 +131,6 @@ LUsolve(Matrix A, int *indexarray, Vector b, Vector x)
     return 0;
 }
 
-/* m_inverse -- returns inverse of A, provided A is not too rank deficient
- *           -- uses LU factorisation */
-#if 0
-Matrix
-m_inverse(Matrix A, Matrix out)
-{
-    int *indexarray = NewAtom_N(int, A->dim);
-    Matrix A1 = new_matrix(A->dim);
-    m_copy(A, A1);
-    LUfactor(A1, indexarray);
-    return LUinverse(A1, indexarray, out);
-}
-#endif				/* 0 */
-
 Matrix
 LUinverse(Matrix A, int *indexarray, Matrix out)
 {
