@@ -1,4 +1,4 @@
-/* $Id: matrix.h,v 1.7 2002/07/18 14:59:02 ukai Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * matrix.h, matrix.c: Liner equation solver using LU decomposition.
  * 
@@ -41,7 +41,7 @@ typedef struct vector *Vector;
 
 #define m_entry(m,i,j) (M_VAL(m,i,j))
 #define v_entry(v,i) (V_VAL(v,i))
-#define m_copy(m1,m2) (bcopy((m1)->me,(m2)->me,(m1)->dim*(m1)->dim*sizeof(double)))
+#define m_copy(m1,m2) (memmove((m2)->me,(m1)->me,(m1)->dim*(m1)->dim*sizeof(double)))
 #define v_free(v) ((v)=NULL)
 #define m_free(m) ((m)=NULL)
 #define px_free(px) ((px)=NULL)
@@ -59,7 +59,7 @@ typedef int PERM;
 extern int LUfactor(Matrix, int *);
 extern Matrix m_inverse(Matrix, Matrix);
 extern Matrix LUinverse(Matrix, int *, Matrix);
-extern int LUsolve(Matrix, int *, Vector, Vector);
+extern int LUsolve(Matrix, const int *, Vector, Vector);
 extern int Lsolve(Matrix, Vector, Vector, double);
 extern int Usolve(Matrix, Vector, Vector, double);
 extern Matrix new_matrix(int);
