@@ -1,4 +1,4 @@
-/* $Id: mktable.c,v 1.15 2007/05/31 01:19:50 inu Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 #include <stdio.h>
 #include <stdlib.h>
 #include "myctype.h"
@@ -6,7 +6,7 @@
 #include "hash.h"
 #include "myctype.h"
 #include "Str.h"
-#include <gc.h>
+#include <gc/gc.h>
 
 /* *INDENT-OFF* */
 defhash(HashItem_ss *, int, hss_i)
@@ -26,7 +26,7 @@ defhashfunc(HashItem_ss *, int, hss_i)
 /* *INDENT-ON* */
 
 int
-main(int argc, char *argv[], char **envp)
+main(int argc, char *argv[])
 {
     FILE *f;
     Hash_ss *hash;
@@ -102,7 +102,7 @@ main(int argc, char *argv[], char **envp)
     }
     printf("static HashItem_si MyHashItem[] = {\n");
     for (i = 0; i < j; i++) {
-	printf("  /* %d */ {\"%s\",%s,", i,
+	printf("    /* %d */ {\"%s\", %s, ", i,
 	       hashitems[i]->key, hashitems[i]->value);
 	if (hashitems[i]->next == NULL) {
 	    printf("NULL},\n");
@@ -116,13 +116,13 @@ main(int argc, char *argv[], char **envp)
 
     for (i = 0; i < hash->size; i++) {
 	if (hash->tab[i])
-	    printf("  &MyHashItem[%d],\n",
+	    printf("    &MyHashItem[%d],\n",
 		   getHash_hss_i(rhash, hash->tab[i], -1));
 	else
-	    printf("  NULL,\n");
+	    printf("    NULL,\n");
     }
     printf("};\n\n");
-    printf("Hash_si %s = {%d, MyHashItemTbl};\n", fbase->ptr, hash->size);
+    printf("Hash_si %s = { %d, MyHashItemTbl };\n", fbase->ptr, hash->size);
 
     exit(0);
 }
