@@ -1,15 +1,21 @@
-/* $Id: rc.c,v 1.116 2010/08/20 09:47:09 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * Initialization file etc.
  */
 #include "fm.h"
 #include "myctype.h"
 #include "proto.h"
+#include "version.h"
+
 #include <stdio.h>
 #include <errno.h>
 #include "parsetag.h"
 #include "local.h"
+#include "regex.h"
 #include <stdlib.h>
+#include <stddef.h>
+#include "rc.h"
+#include "cookie.h"
 
 struct param_ptr {
     char *name;
@@ -60,8 +66,10 @@ static int OptionEncode = FALSE;
 #define CMT_HELPER	 N_("External Viewer Setup")
 #define CMT_TABSTOP      N_("Tab width in characters")
 #define CMT_INDENT_INCR  N_("Indent for HTML rendering")
+#define CMT_MAXCOLS	 N_("Max chars per line (0: unset) Needs restart!")
 #define CMT_PIXEL_PER_CHAR N_("Number of pixels per character (4.0...32.0)")
 #define CMT_PIXEL_PER_LINE N_("Number of pixels per line (4.0...64.0)")
+#define CMT_MESSAGE_DELAY N_("Number of seconds to display user messages")
 #define CMT_PAGERLINE    N_("Number of remembered lines when used as a pager")
 #define CMT_HISTORY	 N_("Use URL history")
 #define CMT_HISTSIZE     N_("Number of remembered URL")
@@ -73,8 +81,10 @@ static int OptionEncode = FALSE;
 #define CMT_OPEN_TAB_DL_LIST N_("Open download list panel on new tab")
 #define CMT_DISPLINK     N_("Display link URL automatically")
 #define CMT_DISPLINKNUMBER N_("Display link numbers")
+#define CMT_ZEROBASEDLINKNO N_("Use 0 as first link number")
 #define CMT_DECODE_URL   N_("Display decoded URL")
-#define CMT_DISPLINEINFO N_("Display current line number")
+#define CMT_DISPLINEINFO N_("Display current line information")
+#define CMT_DISP_COLUMN_NUMBER N_("Display column number in line information")
 #define CMT_DISP_IMAGE   N_("Display inline images")
 #define CMT_PSEUDO_INLINES N_("Display pseudo-ALTs for inline images with no ALT or TITLE string")
 #ifdef USE_IMAGE
@@ -84,13 +94,18 @@ static int OptionEncode = FALSE;
 #define CMT_IMAGE_SCALE  N_("Scale of image (%)")
 #define CMT_IMGDISPLAY   N_("External command to display image")
 #define CMT_IMAGE_MAP_LIST N_("Use link list of image map")
+#define CMT_INLINE_IMG_PROTOCOL N_("Inline image display method")
 #endif
 #define CMT_MULTICOL     N_("Display file names in multi-column format")
 #define CMT_ALT_ENTITY   N_("Use ASCII equivalents to display entities")
 #define CMT_GRAPHIC_CHAR N_("Character type for border of table and menu")
+#define CMT_DISP_BORDERS N_("Display table borders, ignore value of BORDER")
+#define CMT_DISABLE_CENTER N_("Disable center alignment")
 #define CMT_FOLD_TEXTAREA N_("Fold lines in TEXTAREA")
+#define CMT_FOLD_PRE N_("Fold lines in PRE")
 #define CMT_DISP_INS_DEL N_("Display INS, DEL, S and STRIKE element")
 #define CMT_COLOR        N_("Display with color")
+#define CMT_HINTENSITY_COLOR N_("Use high-intensity colors")
 #define CMT_B_COLOR      N_("Color of normal character")
 #define CMT_A_COLOR      N_("Color of anchor")
 #define CMT_I_COLOR      N_("Color of image link")
@@ -122,12 +137,16 @@ static int OptionEncode = FALSE;
 #define CMT_DROOT       N_("Directory corresponding to / (document root)")
 #define CMT_PDROOT      N_("Directory corresponding to /~user")
 #define CMT_CGIBIN      N_("Directory corresponding to /cgi-bin")
-#define CMT_CONFIRM_QQ  N_("Confirm when quitting with q")
-#define CMT_CLOSE_TAB_BACK N_("Close tab if buffer is last when back")
+#define CMT_TMP         N_("Directory for temporary files")
+#define CMT_DL_DIR	N_("Default directory for downloads")
+#define CMT_CONFIRM_QQ  N_("Confirm when calling quit")
+#define CMT_EXIT_ON_LAST  N_("Close w3m/tab if buffer is last when going back")
+#define CMT_CLOSE_TAB_BACK N_("Close tab if buffer is last when going back")
 #ifdef USE_MARK
 #define CMT_USE_MARK	N_("Enable mark operations")
 #endif
 #define CMT_EMACS_LIKE_LINEEDIT	N_("Enable Emacs-style line editing")
+#define CMT_SPACE_AUTOCOMPLETE  N_("Space key triggers file completion while editing URLs")
 #define CMT_VI_PREC_NUM	 N_("Enable vi-like numeric prefix")
 #define CMT_LABEL_TOPLINE N_("Move cursor to top line when going to label")
 #define CMT_NEXTPAGE_TOPLINE N_("Move cursor to top line when moving to next page")
@@ -140,12 +159,19 @@ static int OptionEncode = FALSE;
 #define CMT_EDITOR       N_("Editor")
 #define CMT_MAILER       N_("Mailer")
 #define CMT_MAILTO_OPTIONS N_("How to call Mailer for mailto URLs with options")
-#define CMT_EXTBRZ       N_("External Browser")
-#define CMT_EXTBRZ2      N_("Second External Browser")
-#define CMT_EXTBRZ3      N_("Third External Browser")
+#define CMT_EXTBRZ       N_("External browser")
+#define CMT_EXTBRZ2      N_("2nd external browser")
+#define CMT_EXTBRZ3      N_("3rd external browser")
+#define CMT_EXTBRZ4      N_("4th external browser")
+#define CMT_EXTBRZ5      N_("5th external browser")
+#define CMT_EXTBRZ6      N_("6th external browser")
+#define CMT_EXTBRZ7      N_("7th external browser")
+#define CMT_EXTBRZ8      N_("8th external browser")
+#define CMT_EXTBRZ9      N_("9th external browser")
 #define CMT_DISABLE_SECRET_SECURITY_CHECK	N_("Disable secret file security check")
 #define CMT_PASSWDFILE	 N_("Password file")
 #define CMT_PRE_FORM_FILE	N_("File for setting form on loading")
+#define CMT_SITECONF_FILE	N_("File for preferences for each site")
 #define CMT_FTPPASS      N_("Password for anonymous FTP (your mail address)")
 #define CMT_FTPPASS_HOSTNAMEGEN N_("Generate domain part of password for FTP")
 #define CMT_USERAGENT    N_("User-Agent identification string")
@@ -166,10 +192,15 @@ static int OptionEncode = FALSE;
 #ifdef USE_DICT
 #define CMT_USE_DICTCOMMAND  N_("Enable dictionary lookup through CGI")
 #define CMT_DICTCOMMAND  N_("URL of dictionary lookup command")
+#define CMT_DICTPROMPT	N_("Dictionary command prompt")
 #endif				/* USE_DICT */
 #define CMT_IGNORE_NULL_IMG_ALT	N_("Display link name for images lacking ALT")
 #define CMT_IFILE        N_("Index file for directories")
+#ifdef USE_SSL
+#define CMT_RETRY_HTTP   N_("Prepend https:// to URL automatically")
+#else
 #define CMT_RETRY_HTTP   N_("Prepend http:// to URL automatically")
+#endif				/* USE_SSL */
 #define CMT_DEFAULT_URL  N_("Default value for open-URL command")
 #define CMT_DECODE_CTE   N_("Decode Content-Transfer-Encoding when saving")
 #define CMT_PRESERVE_TIMESTAMP N_("Preserve timestamp when saving")
@@ -182,6 +213,7 @@ static int OptionEncode = FALSE;
 #endif				/* USE_MOUSE */
 #define CMT_CLEAR_BUF     N_("Free memory of undisplayed buffers")
 #define CMT_NOSENDREFERER N_("Suppress `Referer:' header")
+#define CMT_CROSSORIGINREFERER N_("Exclude pathname and query string from `Referer:' header when cross domain communication")
 #define CMT_IGNORE_CASE N_("Search case-insensitively")
 #define CMT_USE_LESSOPEN N_("Use LESSOPEN")
 #ifdef USE_SSL
@@ -191,8 +223,13 @@ static int OptionEncode = FALSE;
 #define CMT_SSL_KEY_FILE N_("PEM encoded private key file of client")
 #define CMT_SSL_CA_PATH N_("Path to directory for PEM encoded certificates of CAs")
 #define CMT_SSL_CA_FILE N_("File consisting of PEM encoded certificates of CAs")
+#define CMT_SSL_CA_DEFAULT N_("Use default locations for PEM encoded certificates of CAs")
 #endif				/* USE_SSL_VERIFY */
-#define CMT_SSL_FORBID_METHOD N_("List of forbidden SSL methods (2: SSLv2, 3: SSLv3, t:TLSv1)")
+#define CMT_SSL_FORBID_METHOD N_("List of forbidden SSL methods (2: SSLv2, 3: SSLv3, t: TLSv1.0, 5: TLSv1.1, 6: TLSv1.2, 7: TLSv1.3)")
+#ifdef SSL_CTX_set_min_proto_version
+#define CMT_SSL_MIN_VERSION N_("Minimum SSL version (all, TLSv1.0, TLSv1.1, TLSv1.2, or TLSv1.3)")
+#endif
+#define CMT_SSL_CIPHER N_("SSL ciphers for TLSv1.2 and below (e.g. DEFAULT:@SECLEVEL=2)")
 #endif				/* USE_SSL */
 #ifdef USE_COOKIE
 #define CMT_USECOOKIE   N_("Enable cookie processing")
@@ -205,6 +242,7 @@ static int OptionEncode = FALSE;
 #endif
 #define CMT_FOLLOW_REDIRECTION N_("Number of redirections to follow")
 #define CMT_META_REFRESH N_("Enable processing of meta-refresh tag")
+#define CMT_LOCALHOST_ONLY N_("Restrict connections only to localhost")
 
 #ifdef USE_MIGEMO
 #define CMT_USE_MIGEMO N_("Enable Migemo (Roma-ji search)")
@@ -214,18 +252,18 @@ static int OptionEncode = FALSE;
 #ifdef USE_M17N
 #define CMT_DISPLAY_CHARSET  N_("Display charset")
 #define CMT_DOCUMENT_CHARSET N_("Default document charset")
-#define CMT_AUTO_DETECT      N_("Automatic charset detect when loading")
+#define CMT_AUTO_DETECT      N_("Automatic charset detection when loading")
 #define CMT_SYSTEM_CHARSET   N_("System charset")
 #define CMT_FOLLOW_LOCALE    N_("System charset follows locale(LC_CTYPE)")
 #define CMT_EXT_HALFDUMP     N_("Output halfdump with display charset")
-#define CMT_USE_WIDE         N_("Use multi column characters")
+#define CMT_USE_WIDE         N_("Use multi-column characters")
 #define CMT_USE_COMBINING    N_("Use combining characters")
 #define CMT_EAST_ASIAN_WIDTH N_("Use double width for some Unicode characters")
 #define CMT_USE_LANGUAGE_TAG N_("Use Unicode language tags")
 #define CMT_UCS_CONV         N_("Charset conversion using Unicode map")
 #define CMT_PRE_CONV         N_("Charset conversion when loading")
 #define CMT_SEARCH_CONV      N_("Adjust search string for document charset")
-#define CMT_FIX_WIDTH_CONV   N_("Fix character width when conversion")
+#define CMT_FIX_WIDTH_CONV   N_("Fix character width when converting")
 #define CMT_USE_GB12345_MAP  N_("Use GB 12345 Unicode map instead of GB 2312's")
 #define CMT_USE_JISX0201     N_("Use JIS X 0201 Roman for ISO-2022-JP")
 #define CMT_USE_JISC6226     N_("Use JIS C 6226:1978 for ISO-2022-JP")
@@ -267,17 +305,8 @@ static struct sel_c colorstr[] = {
 };
 #endif				/* USE_COLOR */
 
-#if 1				/* ANSI-C ? */
 #define N_STR(x)	#x
 #define N_S(x)	(x), N_STR(x)
-#else				/* for traditional cpp? */
-static char n_s[][2] = {
-    {'0', 0},
-    {'1', 0},
-    {'2', 0},
-};
-#define N_S(x) (x), n_s[(x)]
-#endif
 
 
 static struct sel_c defaulturls[] = {
@@ -316,9 +345,6 @@ static struct sel_c dnsorders[] = {
 #ifdef USE_COOKIE
 static struct sel_c badcookiestr[] = {
     {N_S(ACCEPT_BAD_COOKIE_DISCARD), N_("discard")},
-#if 0
-    {N_S(ACCEPT_BAD_COOKIE_ACCEPT), N_("accept")},
-#endif
     {N_S(ACCEPT_BAD_COOKIE_ASK), N_("ask")},
     {0, NULL, NULL}
 };
@@ -352,10 +378,22 @@ static struct sel_c graphic_char_str[] = {
     {0, NULL, NULL}
 };
 
+#ifdef USE_IMAGE
+static struct sel_c inlineimgstr[] = {
+    {N_S(INLINE_IMG_NONE), N_("external command")},
+    {N_S(INLINE_IMG_OSC5379), N_("OSC 5379 (mlterm)")},
+    {N_S(INLINE_IMG_SIXEL), N_("sixel (img2sixel)")},
+    {N_S(INLINE_IMG_ITERM2), N_("OSC 1337 (iTerm2)")},
+    {N_S(INLINE_IMG_KITTY), N_("kitty (ImageMagick)")},
+    {0, NULL, NULL}
+};
+#endif				/* USE_IMAGE */
+
 struct param_ptr params1[] = {
     {"tabstop", P_NZINT, PI_TEXT, (void *)&Tabstop, CMT_TABSTOP, NULL},
     {"indent_incr", P_NZINT, PI_TEXT, (void *)&IndentIncr, CMT_INDENT_INCR,
      NULL},
+    {"maxcols", P_INT, PI_TEXT, (void *)&MaxCols, CMT_MAXCOLS, NULL},
     {"pixel_per_char", P_PIXELS, PI_TEXT, (void *)&pixel_per_char,
      CMT_PIXEL_PER_CHAR, NULL},
 #ifdef USE_IMAGE
@@ -375,6 +413,8 @@ struct param_ptr params1[] = {
     {"decode_url", P_INT, PI_ONOFF, (void *)&DecodeURL, CMT_DECODE_URL, NULL},
     {"display_lineinfo", P_INT, PI_ONOFF, (void *)&displayLineInfo,
      CMT_DISPLINEINFO, NULL},
+    {"display_column_number", P_INT, PI_ONOFF, (void *)&displayColumnNumber,
+     CMT_DISP_COLUMN_NUMBER, NULL},
     {"ext_dirlist", P_INT, PI_ONOFF, (void *)&UseExternalDirBuffer,
      CMT_EXT_DIRLIST, NULL},
     {"dirlist_cmd", P_STRING, PI_TEXT, (void *)&DirBufferCommand,
@@ -384,14 +424,21 @@ struct param_ptr params1[] = {
      CMT_USE_DICTCOMMAND, NULL},
     {"dictcommand", P_STRING, PI_TEXT, (void *)&DictCommand,
      CMT_DICTCOMMAND, NULL},
+    {"dictprompt", P_STRING, PI_TEXT, (void *)&DictPrompt, CMT_DICTPROMPT, NULL},
 #endif				/* USE_DICT */
     {"multicol", P_INT, PI_ONOFF, (void *)&multicolList, CMT_MULTICOL, NULL},
     {"alt_entity", P_CHARINT, PI_ONOFF, (void *)&UseAltEntity, CMT_ALT_ENTITY,
      NULL},
     {"graphic_char", P_CHARINT, PI_SEL_C, (void *)&UseGraphicChar,
      CMT_GRAPHIC_CHAR, (void *)graphic_char_str},
+    {"display_borders", P_CHARINT, PI_ONOFF, (void *)&DisplayBorders,
+     CMT_DISP_BORDERS, NULL},
+    {"disable_center", P_CHARINT, PI_ONOFF, (void *)&DisableCenter,
+     CMT_DISABLE_CENTER, NULL},
     {"fold_textarea", P_CHARINT, PI_ONOFF, (void *)&FoldTextarea,
      CMT_FOLD_TEXTAREA, NULL},
+    {"fold_pre", P_CHARINT, PI_ONOFF, (void *)&FoldPre,
+     CMT_FOLD_PRE, NULL},
     {"display_ins_del", P_INT, PI_SEL_C, (void *)&displayInsDel,
      CMT_DISP_INS_DEL, displayinsdel},
     {"ignore_null_img_alt", P_INT, PI_ONOFF, (void *)&ignore_null_img_alt,
@@ -411,6 +458,8 @@ struct param_ptr params1[] = {
      CMT_EXT_IMAGE_VIEWER, NULL},
     {"image_scale", P_SCALE, PI_TEXT, (void *)&image_scale, CMT_IMAGE_SCALE,
      NULL},
+    {"inline_img_protocol", P_INT, PI_SEL_C, (void *)&enable_inline_image,
+     CMT_INLINE_IMG_PROTOCOL, (void *)inlineimgstr},
     {"imgdisplay", P_STRING, PI_TEXT, (void *)&Imgdisplay, CMT_IMGDISPLAY,
      NULL},
     {"image_map_list", P_INT, PI_ONOFF, (void *)&image_map_list,
@@ -420,6 +469,8 @@ struct param_ptr params1[] = {
     {"show_lnum", P_INT, PI_ONOFF, (void *)&showLineNum, CMT_SHOW_NUM, NULL},
     {"show_srch_str", P_INT, PI_ONOFF, (void *)&show_srch_str,
      CMT_SHOW_SRCH_STR, NULL},
+    {"message_delay", P_NZINT, PI_TEXT, (void *)&MessageDelay, CMT_MESSAGE_DELAY,
+     NULL},
     {"label_topline", P_INT, PI_ONOFF, (void *)&label_topline,
      CMT_LABEL_TOPLINE, NULL},
     {"nextpage_topline", P_INT, PI_ONOFF, (void *)&nextpage_topline,
@@ -430,6 +481,7 @@ struct param_ptr params1[] = {
 #ifdef USE_COLOR
 struct param_ptr params2[] = {
     {"color", P_INT, PI_ONOFF, (void *)&useColor, CMT_COLOR, NULL},
+    {"high-intensity", P_INT, PI_ONOFF, (void *)&highIntensityColors, CMT_HINTENSITY_COLOR, NULL},
     {"basic_color", P_COLOR, PI_SEL_C, (void *)&basic_color, CMT_B_COLOR,
      (void *)colorstr},
     {"anchor_color", P_COLOR, PI_SEL_C, (void *)&anchor_color, CMT_A_COLOR,
@@ -466,6 +518,8 @@ struct param_ptr params3[] = {
 #endif				/* USE_HISTORY */
     {"confirm_qq", P_INT, PI_ONOFF, (void *)&confirm_on_quit, CMT_CONFIRM_QQ,
      NULL},
+    {"exit_on_last", P_INT, PI_ONOFF, (void *)&exit_on_last, CMT_EXIT_ON_LAST,
+     NULL},
     {"close_tab_back", P_INT, PI_ONOFF, (void *)&close_tab_back,
      CMT_CLOSE_TAB_BACK, NULL},
 #ifdef USE_MARK
@@ -473,6 +527,8 @@ struct param_ptr params3[] = {
 #endif
     {"emacs_like_lineedit", P_INT, PI_ONOFF, (void *)&emacs_like_lineedit,
      CMT_EMACS_LIKE_LINEEDIT, NULL},
+    {"space_autocomplete", P_INT, PI_ONOFF, (void *)&space_autocomplete,
+     CMT_SPACE_AUTOCOMPLETE, NULL},
     {"vi_prec_num", P_INT, PI_ONOFF, (void *)&vi_prec_num, CMT_VI_PREC_NUM,
      NULL},
     {"mark_all_pages", P_INT, PI_ONOFF, (void *)&MarkAllPages,
@@ -541,6 +597,8 @@ struct param_ptr params5[] = {
      (void *)&personal_document_root, CMT_PDROOT, NULL},
     {"cgi_bin", P_STRING, PI_TEXT, (void *)&cgi_bin, CMT_CGIBIN, NULL},
     {"index_file", P_STRING, PI_TEXT, (void *)&index_file, CMT_IFILE, NULL},
+    {"tmp_dir", P_STRING, PI_TEXT, (void *)&param_tmp_dir, CMT_TMP, NULL},
+    {"dl_dir", P_STRING, PI_TEXT, (void *)&param_dl_dir, CMT_DL_DIR, NULL},
     {NULL, 0, 0, NULL, NULL, NULL},
 };
 
@@ -561,6 +619,18 @@ struct param_ptr params6[] = {
      NULL},
     {"extbrowser3", P_STRING, PI_TEXT, (void *)&ExtBrowser3, CMT_EXTBRZ3,
      NULL},
+    {"extbrowser4", P_STRING, PI_TEXT, (void *)&ExtBrowser4, CMT_EXTBRZ4,
+     NULL},
+    {"extbrowser5", P_STRING, PI_TEXT, (void *)&ExtBrowser5, CMT_EXTBRZ5,
+     NULL},
+    {"extbrowser6", P_STRING, PI_TEXT, (void *)&ExtBrowser6, CMT_EXTBRZ6,
+     NULL},
+    {"extbrowser7", P_STRING, PI_TEXT, (void *)&ExtBrowser7, CMT_EXTBRZ7,
+     NULL},
+    {"extbrowser8", P_STRING, PI_TEXT, (void *)&ExtBrowser8, CMT_EXTBRZ8,
+     NULL},
+    {"extbrowser9", P_STRING, PI_TEXT, (void *)&ExtBrowser9, CMT_EXTBRZ9,
+     NULL},
     {"bgextviewer", P_INT, PI_ONOFF, (void *)&BackgroundExtViewer,
      CMT_BGEXTVIEW, NULL},
     {"use_lessopen", P_INT, PI_ONOFF, (void *)&use_lessopen, CMT_USE_LESSOPEN,
@@ -572,6 +642,12 @@ struct param_ptr params6[] = {
 struct param_ptr params7[] = {
     {"ssl_forbid_method", P_STRING, PI_TEXT, (void *)&ssl_forbid_method,
      CMT_SSL_FORBID_METHOD, NULL},
+#ifdef SSL_CTX_set_min_proto_version
+    {"ssl_min_version", P_STRING, PI_TEXT, (void *)&ssl_min_version,
+     CMT_SSL_MIN_VERSION, NULL},
+#endif
+    {"ssl_cipher", P_STRING, PI_TEXT, (void *)&ssl_cipher, CMT_SSL_CIPHER,
+     NULL},
 #ifdef USE_SSL_VERIFY
     {"ssl_verify_server", P_INT, PI_ONOFF, (void *)&ssl_verify_server,
      CMT_SSL_VERIFY_SERVER, NULL},
@@ -583,6 +659,8 @@ struct param_ptr params7[] = {
      NULL},
     {"ssl_ca_file", P_SSLPATH, PI_TEXT, (void *)&ssl_ca_file, CMT_SSL_CA_FILE,
      NULL},
+    {"ssl_ca_default", P_INT, PI_ONOFF, (void *)&ssl_ca_default,
+     CMT_SSL_CA_DEFAULT, NULL},
 #endif				/* USE_SSL_VERIFY */
     {NULL, 0, 0, NULL, NULL, NULL},
 };
@@ -619,9 +697,13 @@ struct param_ptr params9[] = {
      CMT_FTPPASS_HOSTNAMEGEN, NULL},
     {"pre_form_file", P_STRING, PI_TEXT, (void *)&pre_form_file,
      CMT_PRE_FORM_FILE, NULL},
+    {"siteconf_file", P_STRING, PI_TEXT, (void *)&siteconf_file,
+     CMT_SITECONF_FILE, NULL},
     {"user_agent", P_STRING, PI_TEXT, (void *)&UserAgent, CMT_USERAGENT, NULL},
     {"no_referer", P_INT, PI_ONOFF, (void *)&NoSendReferer, CMT_NOSENDREFERER,
      NULL},
+    {"cross_origin_referer", P_INT, PI_ONOFF, (void *)&CrossOriginReferer,
+     CMT_CROSSORIGINREFERER, NULL},
     {"accept_language", P_STRING, PI_TEXT, (void *)&AcceptLang, CMT_ACCEPTLANG,
      NULL},
     {"accept_encoding", P_STRING, PI_TEXT, (void *)&AcceptEncoding,
@@ -639,6 +721,8 @@ struct param_ptr params9[] = {
      CMT_FOLLOW_REDIRECTION, NULL},
     {"meta_refresh", P_CHARINT, PI_ONOFF, (void *)&MetaRefresh,
      CMT_META_REFRESH, NULL},
+    {"localhost_only", P_CHARINT, PI_ONOFF, (void *)&LocalhostOnly,
+     CMT_LOCALHOST_ONLY, NULL},
 #ifdef INET6
     {"dns_order", P_INT, PI_SEL_C, (void *)&DNS_order, CMT_DNS_ORDER,
      (void *)dnsorders},
@@ -710,6 +794,12 @@ struct param_ptr params10[] = {
 };
 #endif
 
+struct param_ptr params11[] = {	/* experimental */
+    {"zero_based_link_no", P_INT, PI_ONOFF, (void *)&zeroBasedLinkNo,
+     CMT_ZEROBASEDLINKNO, NULL},
+    {NULL, 0, 0, NULL, NULL, NULL}
+};
+
 struct param_section sections[] = {
     {N_("Display Settings"), params1},
 #ifdef USE_COLOR
@@ -729,6 +819,7 @@ struct param_section sections[] = {
 #ifdef USE_M17N
     {N_("Charset Settings"), params10},
 #endif
+    {N_("Experimental Features"), params11},
     {NULL, NULL}
 };
 
@@ -741,7 +832,7 @@ compare_table(struct rc_search_table *a, struct rc_search_table *b)
 }
 
 static void
-create_option_search_table()
+create_option_search_table(void)
 {
     int i, j, k;
     int diff1, diff2;
@@ -771,7 +862,7 @@ create_option_search_table()
     qsort(RC_search_table, RC_table_size, sizeof(struct rc_search_table),
 	  (int (*)(const void *, const void *))compare_table);
 
-    diff1 = diff2 = 0;
+    diff2 = 0;
     for (i = 0; i < RC_table_size - 1; i++) {
 	p = RC_search_table[i].param->name;
 	q = RC_search_table[i + 1].param->name;
@@ -785,8 +876,8 @@ create_option_search_table()
     }
 }
 
-struct param_ptr *
-search_param(char *name)
+static struct param_ptr *
+search_param(const char *name)
 {
     size_t b, e, i;
     int cmp;
@@ -829,7 +920,7 @@ void
 show_params(FILE * fp)
 {
     int i, j, l;
-    char *t = NULL;
+    const char *t = "";
     char *cmt;
 
 #ifdef USE_M17N
@@ -906,7 +997,7 @@ show_params(FILE * fp)
 }
 
 int
-str_to_bool(char *value, int old)
+str_to_bool(const char *value, int old)
 {
     if (value == NULL)
 	return 1;
@@ -934,7 +1025,7 @@ str_to_bool(char *value, int old)
 
 #ifdef USE_COLOR
 static int
-str_to_color(char *value)
+str_to_color(const char *value)
 {
     if (value == NULL)
 	return 8;		/* terminal */
@@ -975,7 +1066,7 @@ str_to_color(char *value)
 #endif
 
 static int
-set_param(char *name, char *value)
+set_param(const char *name, char *value)
 {
     struct param_ptr *p;
     double ppc;
@@ -1007,7 +1098,7 @@ set_param(char *name, char *value)
 	*(char *)p->varptr = value[0];
 	break;
     case P_STRING:
-	*(char **)p->varptr = value;
+	*(const char **)p->varptr = value;
 	break;
 #if defined(USE_SSL) && defined(USE_SSL_VERIFY)
     case P_SSLPATH:
@@ -1115,8 +1206,8 @@ interpret_rc(FILE * f)
     }
 }
 
-void
-parse_proxy()
+static void
+parse_proxy(void)
 {
     if (non_null(HTTP_proxy))
 	parseURL(HTTP_proxy, &HTTP_proxy_parsed, NULL);
@@ -1134,20 +1225,6 @@ parse_proxy()
 	set_no_proxy(NO_proxy);
 }
 
-#ifdef USE_COOKIE
-void
-parse_cookie()
-{
-    if (non_null(cookie_reject_domains))
-	Cookie_reject_domains = make_domain_list(cookie_reject_domains);
-    if (non_null(cookie_accept_domains))
-	Cookie_accept_domains = make_domain_list(cookie_accept_domains);
-    if (non_null(cookie_avoid_wrong_number_of_dots))
-	Cookie_avoid_wrong_number_of_dots_domains
-	       	= make_domain_list(cookie_avoid_wrong_number_of_dots);
-}
-#endif
-
 #ifdef __EMX__
 static int
 do_mkdir(const char *dir, long mode)
@@ -1173,9 +1250,60 @@ do_mkdir(const char *dir, long mode)
 #endif				/* not __MINW32_VERSION */
 #endif				/* not __EMX__ */
 
+static int
+do_recursive_mkdir(const char *dir)
+{
+    char *ch, *dircpy, tmp;
+    struct stat st;
+
+    if (*dir == '\0')
+	 return -1;
+
+    dircpy = Strnew_charp(dir)->ptr; 
+    ch = dircpy + 1;
+    do {
+	while (!(*ch == '/' || *ch == '\0')) {
+	    ch++;
+	}
+
+	tmp = *ch;
+	*ch = '\0';
+
+	if (stat(dircpy, &st) < 0) {
+	    if (errno != ENOENT) {	/* no directory */
+		return -1; 
+	    }
+	    if (do_mkdir(dircpy, 0700) < 0) {
+		return -1; 
+	    }
+	    stat(dircpy, &st);
+	}
+	if (!S_ISDIR(st.st_mode)) {
+	    /* not a directory */
+	    return -1; 
+	}
+	if (!(st.st_mode & S_IWUSR)) {
+	    return -1; 
+	}
+
+	*ch = tmp;
+
+    } while (*ch++ != '\0');
+#ifdef HAVE_FACCESSAT
+    if (faccessat(AT_FDCWD, dir, W_OK|X_OK, AT_EACCESS) < 0) {
+	return -1; 
+    }
+#endif
+
+    return 0;
+}
+
+static void loadSiteconf(void);
+
 void
 sync_with_option(void)
 {
+    init_tmp();
     if (PagerMax < LINES)
 	PagerMax = LINES;
     WrapSearch = WrapDefault;
@@ -1192,13 +1320,14 @@ sync_with_option(void)
     init_migemo();
 #endif
 #ifdef USE_IMAGE
-    if (fmInitialized && displayImage)
+    if (fmInitialized && (displayImage || enable_inline_image))
 	initImage();
 #else
     displayImage = FALSE;	/* XXX */
 #endif
     loadPasswd();
     loadPreForm();
+    loadSiteconf();
 
     if (AcceptLang == NULL || *AcceptLang == '\0') {
 	/* TRANSLATORS: 
@@ -1214,6 +1343,9 @@ sync_with_option(void)
 	AcceptMedia = acceptableMimeTypes();
 #ifdef USE_UNICODE
     update_utf8_symbol();
+#endif
+#ifdef USE_M17N
+    wtf_init(DocumentCharset, DisplayCharset);
 #endif
     if (fmInitialized) {
 	initKeymap(FALSE);
@@ -1230,13 +1362,18 @@ void
 init_rc(void)
 {
     int i;
-    struct stat st;
     FILE *f;
 
     if (rc_dir != NULL)
 	goto open_rc;
 
-    rc_dir = expandPath(RC_DIR);
+    rc_dir = allocStr(getenv("W3M_DIR"), -1);
+    if (rc_dir == NULL || *rc_dir == '\0')
+	rc_dir = allocStr(RC_DIR, -1);
+    if (rc_dir == NULL || *rc_dir == '\0')
+	goto rc_dir_err;
+    rc_dir = expandPath(rc_dir);
+
     i = strlen(rc_dir);
     if (i > 1 && rc_dir[i - 1] == '/')
 	rc_dir[i - 1] = '\0';
@@ -1247,33 +1384,13 @@ init_rc(void)
     system_charset_str = display_charset_str;
 #endif
 
-    if (stat(rc_dir, &st) < 0) {
-	if (errno == ENOENT) {	/* no directory */
-	    if (do_mkdir(rc_dir, 0700) < 0) {
-		fprintf(stderr, "Can't create config directory (%s)!", rc_dir);
-		goto rc_dir_err;
-	    }
-	    else {
-		stat(rc_dir, &st);
-	    }
-	}
-	else {
-	    fprintf(stderr, "Can't open config directory (%s)!", rc_dir);
-	    goto rc_dir_err;
-	}
-    }
-    if (!S_ISDIR(st.st_mode)) {
-	/* not a directory */
-	fprintf(stderr, "%s is not a directory!", rc_dir);
-	goto rc_dir_err;
-    }
-    if (!(st.st_mode & S_IWUSR)) {
-	fprintf(stderr, "%s is not writable!", rc_dir);
-	goto rc_dir_err;
-    }
-    no_rc_dir = FALSE;
     tmp_dir = rc_dir;
 
+    if (do_recursive_mkdir(rc_dir) == -1)
+	goto rc_dir_err;
+
+    no_rc_dir = FALSE;
+
     if (config_file == NULL)
 	config_file = rcFile(CONFIG_FILE);
 
@@ -1297,12 +1414,53 @@ init_rc(void)
 
   rc_dir_err:
     no_rc_dir = TRUE;
+    create_option_search_table();
+    goto open_rc;
+}
+
+void
+init_tmp(void)
+{
+    int i;
+
+    if (param_tmp_dir)
+	tmp_dir = param_tmp_dir;
+    if (*tmp_dir == '\0')
+	tmp_dir = rc_dir;
+
+    if (strcmp(tmp_dir, rc_dir) == 0) {
+	if (no_rc_dir)
+	    goto tmp_dir_err;
+	return;
+    }
+
+    tmp_dir = expandPath(tmp_dir);
+    i = strlen(tmp_dir);
+    if (i > 1 && tmp_dir[i - 1] == '/')
+	tmp_dir[i - 1] = '\0';
+    if (do_recursive_mkdir(tmp_dir) == -1)
+	goto tmp_dir_err;
+    return;
+
+  tmp_dir_err:
+#ifdef HAVE_MKDTEMP
+    if (mkd_tmp_dir) {
+	tmp_dir = mkd_tmp_dir;
+	return;
+    }
+#endif
     if (((tmp_dir = getenv("TMPDIR")) == NULL || *tmp_dir == '\0') &&
 	((tmp_dir = getenv("TMP")) == NULL || *tmp_dir == '\0') &&
 	((tmp_dir = getenv("TEMP")) == NULL || *tmp_dir == '\0'))
 	tmp_dir = "/tmp";
-    create_option_search_table();
-    goto open_rc;
+#ifdef HAVE_MKDTEMP
+    tmp_dir = mkdtemp(Strnew_m_charp(tmp_dir, "/w3m-XXXXXX", NULL)->ptr);
+    if (tmp_dir)
+	mkd_tmp_dir = tmp_dir;
+    else
+	tmp_dir = rc_dir;
+#endif
+    return;
 }
 
 
@@ -1349,7 +1507,7 @@ to_str(struct param_ptr *p)
 	return Sprintf("%g", *(double *)p->varptr);
     }
     /* not reached */
-    return NULL;
+    return Strnew();
 }
 
 Buffer *
@@ -1366,7 +1524,7 @@ load_option_panel(void)
     Buffer *buf;
 
     if (optionpanel_str == NULL)
-	optionpanel_str = Sprintf(optionpanel_src1, w3m_version,
+	optionpanel_str = Sprintf(optionpanel_src1, W3M_VERSION,
 			      html_quote(localCookie()->ptr), _(CMT_HELPER));
 #ifdef USE_M17N
 #ifdef ENABLE_NLS
@@ -1483,6 +1641,7 @@ panel_set_option(struct parsed_tagarg *arg)
 {
     FILE *f = NULL;
     char *p;
+    Str s = Strnew(), tmp;
 
     if (config_file == NULL) {
 	disp_message("There's no config file... config not saved", FALSE);
@@ -1498,14 +1657,17 @@ panel_set_option(struct parsed_tagarg *arg)
 	if (arg->value) {
 	    p = conv_to_system(arg->value);
 	    if (set_param(arg->arg, p)) {
-		if (f)
-		    fprintf(f, "%s %s\n", arg->arg, p);
+		tmp = Sprintf("%s %s\n", arg->arg, p);
+		Strcat(tmp, s);
+		s = tmp;
 	    }
 	}
 	arg = arg->next;
     }
-    if (f)
+    if (f) {
+	fputs(s->ptr, f);
 	fclose(f);
+    }
     sync_with_option();
     backBf();
 }
@@ -1524,35 +1686,253 @@ rcFile(char *base)
 }
 
 char *
-auxbinFile(char *base)
+auxbinFile(const char *base)
 {
     return expandPath(Strnew_m_charp(w3m_auxbin_dir(), "/", base, NULL)->ptr);
 }
 
-#if 0				/* not used */
 char *
-libFile(char *base)
-{
-    return expandPath(Strnew_m_charp(w3m_lib_dir(), "/", base, NULL)->ptr);
-}
-#endif
-
-char *
-etcFile(char *base)
+etcFile(const char *base)
 {
     return expandPath(Strnew_m_charp(w3m_etc_dir(), "/", base, NULL)->ptr);
 }
 
 char *
-confFile(char *base)
+confFile(const char *base)
 {
     return expandPath(Strnew_m_charp(w3m_conf_dir(), "/", base, NULL)->ptr);
 }
 
 #ifndef USE_HELP_CGI
 char *
-helpFile(char *base)
+helpFile(const char *base)
 {
     return expandPath(Strnew_m_charp(w3m_help_dir(), "/", base, NULL)->ptr);
 }
 #endif
+
+/* siteconf */
+/*
+ * url "<url>"|/<re-url>/|m@<re-url>@i [exact]
+ * substitute_url "<destination-url>"
+ * url_charset <charset>
+ * no_referer_from on|off
+ * no_referer_to on|off
+ * user_agent "<string>"
+ * 
+ * The last match wins.
+ */
+
+struct siteconf_rec {
+    struct siteconf_rec *next;
+    char *url;
+    Regex *re_url;
+    int url_exact;
+    unsigned char mask[(SCONF_N_FIELD + 7) >> 3];
+
+    char *substitute_url;
+    char *user_agent;
+#ifdef USE_M17N
+    wc_ces url_charset;
+#endif
+    int no_referer_from;
+    int no_referer_to;
+};
+#define SCONF_TEST(ent, f) ((ent)->mask[(f)>>3] & (1U<<((f)&7)))
+#define SCONF_SET(ent, f) ((ent)->mask[(f)>>3] |= (1U<<((f)&7)))
+#define SCONF_CLEAR(ent, f) ((ent)->mask[(f)>>3] &= ~(1U<<((f)&7)))
+
+static struct siteconf_rec *siteconf_head = NULL;
+static struct siteconf_rec *newSiteconfRec(void);
+
+static struct siteconf_rec *
+newSiteconfRec(void)
+{
+    struct siteconf_rec *ent;
+
+    ent = New(struct siteconf_rec);
+    ent->next = NULL;
+    ent->url = NULL;
+    ent->re_url = NULL;
+    ent->url_exact = FALSE;
+    memset(ent->mask, 0, sizeof(ent->mask));
+
+    ent->substitute_url = NULL;
+    ent->user_agent = NULL;
+#ifdef USE_M17N
+    ent->url_charset = 0;
+#endif
+    return ent;
+}
+
+static void
+loadSiteconf(void)
+{
+    char *efname;
+    FILE *fp;
+    Str line;
+    struct siteconf_rec *ent = NULL;
+
+    siteconf_head = NULL;
+    if (!siteconf_file)
+	return;
+    if ((efname = expandPath(siteconf_file)) == NULL)
+	return;
+    fp = fopen(efname, "r");
+    if (fp == NULL)
+	return;
+    while (line = Strfgets(fp), line->length > 0) {
+	char *p, *s;
+
+	Strchop(line);
+	p = line->ptr;
+	SKIP_BLANKS(p);
+	if (*p == '#' || *p == '\0')
+	    continue;
+	s = getWord(&p);
+
+	/* The "url" begins a new record. */
+	if (strcmp(s, "url") == 0) {
+	    char *url, *opt;
+	    struct siteconf_rec *newent;
+
+	    /* First, register the current record. */
+	    if (ent) {
+		ent->next = siteconf_head;
+		siteconf_head = ent;
+		ent = NULL;
+	    }
+
+	    /* Second, create a new record. */
+	    newent = newSiteconfRec();
+	    url = getRegexWord(&p, &newent->re_url);
+	    opt = getWord(&p);
+	    SKIP_BLANKS(p);
+	    if (!newent->re_url) {
+		ParsedURL pu;
+		if (!url || !*url)
+		    continue;
+		parseURL2(url, &pu, NULL);
+		newent->url = parsedURL2Str(&pu)->ptr;
+	    }
+	    /* If we have an extra or unknown option, ignore this record
+	     * for future extensions. */
+	    if (strcmp(opt, "exact") == 0) {
+		newent->url_exact = TRUE;
+	    }
+	    else if (*opt != 0)
+		    continue;
+	    if (*p)
+		continue;
+	    ent = newent;
+	    continue;
+	}
+
+	/* If the current record is broken, skip to the next "url". */
+	if (!ent)
+	    continue;
+
+	/* Fill the new record. */
+	if (strcmp(s, "substitute_url") == 0) {
+	    ent->substitute_url = getQWord(&p);
+	    SCONF_SET(ent, SCONF_SUBSTITUTE_URL);
+	}
+	if (strcmp(s, "user_agent") == 0) {
+	    ent->user_agent = getQWord(&p);
+	    SCONF_SET(ent, SCONF_USER_AGENT);
+	}
+#ifdef USE_M17N
+	else if (strcmp(s, "url_charset") == 0) {
+	    char *charset = getWord(&p);
+	    ent->url_charset = (charset && *charset) ?
+		wc_charset_to_ces(charset) : 0;
+	    SCONF_SET(ent, SCONF_URL_CHARSET);
+	}
+#endif /* USE_M17N */
+	else if (strcmp(s, "no_referer_from") == 0) {
+	    ent->no_referer_from = str_to_bool(getWord(&p), 0);
+	    SCONF_SET(ent, SCONF_NO_REFERER_FROM);
+	}
+	else if (strcmp(s, "no_referer_to") == 0) {
+	    ent->no_referer_to = str_to_bool(getWord(&p), 0);
+	    SCONF_SET(ent, SCONF_NO_REFERER_TO);
+	}
+    }
+    if (ent) {
+	ent->next = siteconf_head;
+	siteconf_head = ent;
+	ent = NULL;
+    }
+    fclose(fp);
+}
+
+void *
+querySiteconf(ParsedURL *query_pu, int field)
+{
+    struct siteconf_rec *ent;
+    Str u;
+    const char *firstp, *lastp;
+
+    if (field < 0 || field >= SCONF_N_FIELD)
+	return NULL;
+    if (!query_pu || IS_EMPTY_PARSED_URL(query_pu))
+	return NULL;
+    u = parsedURL2Str(query_pu);
+    if (u->length == 0)
+	return NULL;
+
+    for (ent = siteconf_head; ent; ent = ent->next) {
+	if (!SCONF_TEST(ent, field))
+	    continue;
+	if (ent->re_url) {
+	    if (RegexMatch(ent->re_url, u->ptr, u->length, 1)) {
+		MatchedPosition(ent->re_url, &firstp, &lastp);
+		if (!ent->url_exact)
+		    goto url_found;
+		if (firstp != u->ptr || lastp == firstp)
+		    continue;
+		if (*lastp == 0 || *lastp == '?' || *(lastp - 1) == '?' ||
+		    *lastp == '#' || *(lastp - 1) == '#')
+		    goto url_found;
+	    }
+	} else {
+	    int matchlen = strmatchlen(ent->url, u->ptr, u->length);
+	    if (matchlen == 0 || ent->url[matchlen] != 0)
+		continue;
+	    firstp = u->ptr;
+	    lastp = u->ptr + matchlen;
+	    if (*lastp == 0 || *lastp == '?' || *(lastp - 1) == '?' ||
+		*lastp == '#' || *(lastp - 1) == '#')
+		goto url_found;
+	    if (!ent->url_exact && (*lastp == '/' || *(lastp - 1) == '/'))
+		goto url_found;
+	}
+    }
+    return NULL;
+
+url_found:
+    switch (field) {
+    case SCONF_SUBSTITUTE_URL:
+	if (ent->substitute_url && *ent->substitute_url) {
+	    Str tmp = Strnew_charp_n(u->ptr, firstp - u->ptr);
+	    Strcat_charp(tmp, ent->substitute_url);
+	    Strcat_charp(tmp, lastp);
+	    return tmp->ptr;
+	}
+	return NULL;
+    case SCONF_USER_AGENT:
+	if (ent->user_agent && *ent->user_agent) {
+	    return ent->user_agent;
+	}
+	return NULL;
+#ifdef USE_M17N
+    case SCONF_URL_CHARSET:
+	return &ent->url_charset;
+#endif
+    case SCONF_NO_REFERER_FROM:
+	return &ent->no_referer_from;
+    case SCONF_NO_REFERER_TO:
+	return &ent->no_referer_to;
+    }
+    return NULL;
+}
