@@ -1,4 +1,4 @@
-/* $Id: terms.c,v 1.63 2010/08/20 09:34:47 htrb Exp $ */
+/* vi: set sw=4 ts=8 ai sm noet : */
 /* 
  * An original curses library for EUC-kanji by Akinori ITO,     December 1989
  * revised by Akinori ITO, January 1995
@@ -6,12 +6,14 @@
 #include <stdio.h>
 #include <signal.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <sys/time.h>
 #include <unistd.h>
 #include "config.h"
 #include <string.h>
+#include <sys/wait.h>
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
@@ -25,6 +27,7 @@
 #include <gpm.h>
 #endif				/* USE_GPM */
 #ifdef USE_SYSMOUSE
+static int is_xterm = 0;
 #include <osreldate.h>
 #if (__FreeBSD_version >= 400017) || (__FreeBSD_kernel_version >= 400017)
 #include <sys/consio.h>
@@ -37,13 +40,11 @@ static int cwidth = 8, cheight = 16;
 static int xpix, ypix, nbs, obs = 0;
 #endif				/* use_SYSMOUSE */
 
-static int is_xterm = 0;
-
-void mouse_init(), mouse_end();
+static void mouse_init(void);
 int mouseActive = 0;
 #endif				/* USE_MOUSE */
 
-static char *title_str = NULL;
+static const char *title_str;
 
 static int tty;
 
@@ -64,10 +65,6 @@ static int isWinConsole = 0;
 #define TERM_CYGWIN_RESERVE_IME 2
 static int isLocalConsole = 0;
 
-#if CYGWIN_VERSION_DLL_MAJOR < 1005 && defined(USE_MOUSE)
-int cygwin_mouse_btn_swapped = 0;
-#endif
-
 #if defined(SUPPORT_WIN9X_CONSOLE_MBCS)
 static HANDLE hConIn = INVALID_HANDLE_VALUE;
 static int isWin95 = 0;
@@ -157,8 +154,6 @@ read_win32_console_input(void)
 static int
 read_win32_console(char *s, int n)
 {
-    KEY_EVENT_RECORD *ker;
-
     if (hConIn == INVALID_HANDLE_VALUE)
 	return read(tty, s, n);
 
@@ -204,28 +199,13 @@ GetConsoleHwnd(void)
     return (hwndFound);
 }
 
-#if CYGWIN_VERSION_DLL_MAJOR < 1005 && defined(USE_MOUSE)
-static unsigned long
-cygwin_version(void)
-{
-    struct per_process *p;
-
-    p = (struct per_process *)cygwin_internal(CW_USER_DATA);
-    if (p != NULL) {
-	return (p->dll_major * 1000) + p->dll_minor;
-    }
-    return 0;
-}
-#endif
-
 static void
 check_cygwin_console(void)
 {
     char *term = getenv("TERM");
+    char *ctype;
     HANDLE hWnd;
 
-    if (term == NULL)
-	term = DEFAULT_TERM;
     if (term && strncmp(term, "cygwin", 6) == 0) {
 	isWinConsole = TERM_CYGWIN;
     }
@@ -236,7 +216,9 @@ check_cygwin_console(void)
 		isLocalConsole = 1;
 	    }
 	}
-	if (strncmp(getenv("LANG"), "ja", 2) == 0) {
+	if (((ctype = getenv("LC_ALL")) ||
+	     (ctype = getenv("LC_CTYPE")) ||
+	     (ctype = getenv("LANG"))) && strncmp(ctype, "ja", 2) == 0) {
 	    isWinConsole = TERM_CYGWIN_RESERVE_IME;
 	}
 #ifdef SUPPORT_WIN9X_CONSOLE_MBCS
@@ -246,19 +228,10 @@ check_cygwin_console(void)
 	}
 #endif
     }
-#if CYGWIN_VERSION_DLL_MAJOR < 1005 && defined(USE_MOUSE)
-    if (cygwin_version() <= 1003015) {
-	/* cygwin DLL 1.3.15 or earler */
-	cygwin_mouse_btn_swapped = 1;
-    }
-#endif
 }
 #endif				/* __CYGWIN__ */
 
-char *getenv(const char *);
-MySignalHandler reset_exit(SIGNAL_ARG), reset_error_exit(SIGNAL_ARG), error_dump(SIGNAL_ARG);
-void setlinescols(void);
-void flush_tty();
+void reset_exit(SIGNAL_ARG), reset_error_exit(SIGNAL_ARG), error_dump(SIGNAL_ARG);
 
 #ifndef SIGIOT
 #define SIGIOT SIGABRT
@@ -327,9 +300,6 @@ typedef struct termios TerminalMode;
 char *ttyname(int);
 #endif /* __MINGW32_VERSION */
 
-#define MAX_LINE        200
-#define MAX_COLUMN      400
-
 /* Screen properties */
 #define S_SCREENPROP    0x0f
 #define S_NORMAL        0x00
@@ -351,7 +321,7 @@ char *ttyname(int);
 #define SETCHMODE(var,mode)	((var) = (((var)&~C_WHICHCHAR) | mode))
 #ifdef USE_M17N
 #define SETCH(var,ch,len)	((var) = New_Reuse(char, (var), (len) + 1), \
-				strncpy((var), (ch), (len)), (var)[len] = '\0')
+				strncpy((var), (ch), (len + 1)))
 #else
 #define SETCH(var,ch,len)	((var) = (ch))
 #endif
@@ -428,6 +398,7 @@ char *T_cd, *T_ce, *T_kr, *T_kl, *T_cr, *T_bt, *T_ta, *T_sc, *T_rc,
     *T_ti, *T_te, *T_nd, *T_as, *T_ae, *T_eA, *T_ac, *T_op;
 
 int LINES, COLS;
+int opt_cols;
 #if defined(__CYGWIN__)
 int LASTLINE;
 #endif				/* defined(__CYGWIN__) */
@@ -447,10 +418,7 @@ extern int tgetflag(char *);
 extern char *tgetstr(char *, char **);
 extern char *tgoto(char *, int, int);
 extern int tputs(char *, int, int (*)(char));
-void clear(), wrap(), touch_line(), touch_column(int);
-#if 0
-void need_clrtoeol(void);
-#endif
+void wrap(void), touch_line(void), touch_column(int);
 void clrtoeol(void);		/* conflicts with curs_clear(3)? */
 
 static int write1(char);
@@ -463,45 +431,406 @@ writestr(char *s)
 
 #define MOVE(line,column)       writestr(tgoto(T_cm,column,line));
 
-#ifdef USE_MOUSE
-#define W3M_TERM_INFO(name, title, mouse)	name, title, mouse
-#define NEED_XTERM_ON   (1)
-#define NEED_XTERM_OFF  (1<<1)
-#ifdef __CYGWIN__
-#define NEED_CYGWIN_ON  (1<<2)
-#define NEED_CYGWIN_OFF (1<<3)
-#endif
-#else
-#define W3M_TERM_INFO(name, title, mouse)	name, title
+#ifdef USE_IMAGE
+/*
+ * From the Kitty spec,
+ * <https://sw.kovidgoyal.net/kitty/graphics-protocol/#remote-client>:
+ *
+ * > The pixel data must first be base64 encoded then chunked up into
+ * > chunks no larger than 4096 bytes.
+ *
+ * base64 inflates the size by 4/3, hence we read 3072.
+ *
+ * For simplicity we use the same size everywhere.
+ */
+#define IMG_BUF_SZ 3072
+
+void
+put_image_osc5379(char *url, int x, int y, int w, int h, int sx, int sy, int sw, int sh)
+{
+    Str buf;
+    char *size ;
+
+    if (w > 0 && h > 0)
+	size = Sprintf("%dx%d",w,h)->ptr;
+    else
+	size = "";
+
+    MOVE(y,x);
+    buf = Sprintf("\x1b]5379;show_picture %s %s %dx%d+%d+%d\x07",url,size,sw,sh,sx,sy);
+    writestr(buf->ptr);
+    MOVE(Currentbuf->cursorY,Currentbuf->cursorX);
+}
+
+
+void
+put_image_iterm2(char *url, int x, int y, int w, int h)
+{
+    Str buf;
+    char cbuf[IMG_BUF_SZ];
+    FILE *fp;
+    int n;
+    struct stat st;
+
+    if (stat(url, &st))
+	return;
+
+    if (!(fp = fopen(url, "r")))
+	return;
+
+    buf = Sprintf("\x1b]1337;"
+      "File="
+      "name=%s;"
+      "size=%d;"
+      "width=%d;"
+      "height=%d;"
+      "preserveAspectRatio=0;"
+      "inline=1"
+      ":", url, (int)st.st_size, w, h);
+
+    MOVE(y,x);
+
+    writestr(buf->ptr);
+
+    while ((n = fread(cbuf, 1, IMG_BUF_SZ, fp)))
+	writestr(base64_encode(cbuf, n)->ptr);
+
+    fclose(fp);
+    writestr("\a");
+    MOVE(Currentbuf->cursorY,Currentbuf->cursorX);
+}
+
+void ttymode_set(int mode, int imode);
+void ttymode_reset(int mode, int imode);
+
+void
+put_image_kitty(char *url, int x, int y, int w, int h, int sx, int sy, int sw,
+    int sh, int cols, int rows)
+{
+    Str buf, base64;
+    char *tmpf;
+    char cbuf[IMG_BUF_SZ];
+    char *argv[4];
+    const char *type;
+    FILE *fp;
+    int n, m, is_anim;
+    struct stat st;
+    pid_t pid;
+    void (*volatile previntr) (SIGNAL_ARG);
+    void (*volatile prevquit) (SIGNAL_ARG);
+    void (*volatile prevstop) (SIGNAL_ARG);
+
+    type = guessContentType(url);
+
+    /* convert to PNG, so that we transfer as little data as possible. */
+    if (!type || strcasecmp(type, "image/png")) {
+	tmpf = Sprintf("%s/%s.png", tmp_dir, mybasename(url))->ptr;
+	is_anim = type && !strcasecmp(type, "image/gif");
+
+	/* convert only if png doesn't exist yet. */
+	if (stat(tmpf, &st)) {
+	    if (stat(url, &st))
+		return;
+
+	    flush_tty();
+
+	    previntr = mySignal(SIGINT, SIG_IGN);
+	    prevquit = mySignal(SIGQUIT, SIG_IGN);
+	    prevstop = mySignal(SIGTSTP, SIG_IGN);
+
+	    if ((pid = fork()) == 0) {
+		close(STDERR_FILENO);	/* Don't output error message. */
+		ttymode_set(ISIG, 0);
+
+		if (!(argv[0] = getenv("W3M_KITTY_TO_PNG")))
+		    argv[0] = "convert";
+
+		if (is_anim) {
+		    buf = Strnew_charp(url);
+		    Strcat_charp(buf, "[0]");
+		    argv[1] = buf->ptr;
+		} else {
+		    argv[1] = url;
+		}
+		argv[2] = tmpf;
+		argv[3] = NULL;
+		execvp(argv[0],argv);
+		exit(0);
+	    }
+	    else if (pid > 0) {
+		waitpid(pid, &n, 0);
+		ttymode_reset(ISIG, 0);
+		mySignal(SIGINT, previntr);
+		mySignal(SIGQUIT, prevquit);
+		mySignal(SIGTSTP, prevstop);
+	    }
+
+	    pushText(fileToDelete, tmpf);
+	}
+	url = tmpf;
+    }
+
+    if (!(fp = fopen(url, "r")))
+	return;
+
+    MOVE(y, x);
+
+    n = fread(cbuf, 1, IMG_BUF_SZ, fp);
+    base64 = base64_encode(cbuf, n);
+
+    m = n == IMG_BUF_SZ; /* m=1 -> has more, m=0 -> finished */
+    /* 100 is format = PNG in Kitty */
+    buf = Sprintf("\x1b_Gf=100,a=T,s=%d,v=%d,m=%d,x=%d,y=%d,w=%d,h=%d,c=%d,r=%d"
+	";%s\x1b\\", w, h, m, sx, sy, sw, sh, cols, rows, base64->ptr);
+    writestr(buf->ptr);
+
+    while ((n = fread(cbuf, 1, sizeof cbuf, fp))) {
+	m = n == IMG_BUF_SZ;
+	base64 = base64_encode(cbuf, n);
+	buf = Sprintf("\x1b_Gm=%d;%s\x1b\\", m, base64->ptr);
+	writestr(buf->ptr);
+    }
+
+    fclose(fp);
+    MOVE(Currentbuf->cursorY, Currentbuf->cursorX);
+}
+
+static void
+save_gif(const char *path, u_char *header, size_t  header_size, u_char *body, size_t body_size)
+{
+    int	fd;
+
+    if ((fd = open(path, O_WRONLY|O_CREAT, 0600)) >= 0) {
+	write(fd, header, header_size) ;
+	write(fd, body, body_size) ;
+	write(fd, "\x3b" , 1) ;
+	close(fd) ;
+    }
+}
+
+static u_char *
+skip_gif_header(u_char *p)
+{
+    /* Header */
+    p += 10;
+
+    if (*(p) & 0x80) {
+	p += (3 * (2 << ((*p) & 0x7)));
+    }
+    p += 3;
+
+    return p;
+}
+
+static Str
+save_first_animation_frame(const char *path)
+{
+    int	fd;
+    struct stat	st;
+    u_char *header;
+    size_t header_size;
+    u_char *body;
+    u_char *p;
+    ssize_t len;
+    Str new_path;
+
+    new_path = Strnew_charp(path);
+    Strcat_charp(new_path, "-1");
+    if (stat(new_path->ptr, &st) == 0) {
+	return new_path;
+    }
+
+    if ((fd = open( path, O_RDONLY)) < 0) {
+	return NULL;
+    }
+
+    if (fstat( fd, &st) != 0 || ! (header = GC_malloc( st.st_size))){
+	close( fd);
+	return NULL;
+    }
+
+    len = read(fd, header, st.st_size);
+    close(fd);
+
+    /* Header */
+
+    if (len != st.st_size || strncmp((char *)header, "GIF89a", 6) != 0) {
+	return NULL;
+    }
+
+    p = skip_gif_header(header);
+    header_size = p - header;
+
+    /* Application Extension */
+    if (p[0] == 0x21 && p[1] == 0xff) {
+	p += 19;
+    }
+
+    /* Other blocks */
+    body = NULL;
+    while (p + 2 < header + st.st_size) {
+	if (*(p++) == 0x21 && *(p++) == 0xf9 && *(p++) == 0x04) {
+	    if( body) {
+		/* Graphic Control Extension */
+		save_gif(new_path->ptr, header, header_size, body, p - 3 - body);
+		return new_path;
+	    }
+	    else {
+		/* skip the first frame. */
+	    }
+	    body = p - 3;
+	}
+    }
+
+    return NULL;
+}
+
+void
+put_image_sixel(char *url, int x, int y, int w, int h, int sx, int sy, int sw, int sh, int n_terminal_image)
+{
+    pid_t pid;
+    int do_anim;
+    void (*volatile previntr) (SIGNAL_ARG);
+    void (*volatile prevquit) (SIGNAL_ARG);
+    void (*volatile prevstop) (SIGNAL_ARG);
+
+    MOVE(y,x);
+    flush_tty();
+
+    do_anim = (n_terminal_image == 1 && x == 0 && y == 0 && sx == 0 && sy == 0);
+
+    previntr = mySignal(SIGINT, SIG_IGN);
+    prevquit = mySignal(SIGQUIT, SIG_IGN);
+    prevstop = mySignal(SIGTSTP, SIG_IGN);
+
+    if ((pid = fork()) == 0) {
+	char *env;
+	int n = 0;
+	char *argv[20];
+	char digit[2][11+1];
+	char clip[44+3+1];
+	Str str_url;
+
+	close(STDERR_FILENO);	/* Don't output error message. */
+	if (do_anim) {
+	    writestr("\x1b[?80h");
+	}
+	else if (!strstr(url, "://") && strcmp(url+strlen(url)-4, ".gif") == 0 &&
+                 (str_url = save_first_animation_frame(url))) {
+	    url = str_url->ptr;
+	}
+	ttymode_set(ISIG, 0);
+
+	if ((env = getenv("W3M_IMG2SIXEL"))) {
+	    char *p;
+	    env = Strnew_charp(env)->ptr;
+	    while (n < 8 && (p = strchr(env, ' '))) {
+		*p = '\0';
+		if (*env != '\0') {
+		    argv[n++] = env;
+		}
+		env = p+1;
+	    }
+	    if (*env != '\0') {
+		argv[n++] = env;
+	    }
+	}
+	else {
+		argv[n++] = "img2sixel";
+	}
+	argv[n++] = "-l";
+	argv[n++] = do_anim ? "auto" : "disable";
+	argv[n++] = "-w";
+	sprintf(digit[0], "%d", w);
+	argv[n++] = digit[0];
+	argv[n++] = "-h";
+	sprintf(digit[1], "%d", h);
+	argv[n++] = digit[1];
+	argv[n++] = "-c";
+	sprintf(clip, "%dx%d+%d+%d", sw, sh, sx, sy);
+	argv[n++] = clip;
+	argv[n++] = url;
+	if (getenv("TERM") && strcmp(getenv("TERM"), "screen") == 0 &&
+	    (!getenv("SCREEN_VARIANT") || strcmp(getenv("SCREEN_VARIANT"), "sixel") != 0)) {
+	    argv[n++] = "-P";
+	}
+	argv[n++] = NULL;
+	execvp(argv[0],argv);
+	exit(0);
+    }
+    else if (pid > 0) {
+	int status;
+	waitpid(pid, &status, 0);
+	ttymode_reset(ISIG, 0);
+	mySignal(SIGINT, previntr);
+	mySignal(SIGQUIT, prevquit);
+	mySignal(SIGTSTP, prevstop);
+	if (do_anim) {
+	    writestr("\x1b[?80l");
+	}
+    }
+
+    MOVE(Currentbuf->cursorY,Currentbuf->cursorX);
+}
+
+int
+get_pixel_per_cell(int *ppc, int *ppl)
+{
+    fd_set  rfd;
+    struct timeval tval;
+    char buf[100];
+    char *p;
+    ssize_t len;
+    ssize_t left;
+    int wp,hp,wc,hc;
+    int i;
+
+#ifdef  TIOCGWINSZ
+    struct winsize ws;
+    if (ioctl(tty, TIOCGWINSZ, &ws) == 0 && ws.ws_ypixel > 0 && ws.ws_row > 0 &&
+        ws.ws_xpixel > 0 && ws.ws_col > 0) {
+	*ppc = ws.ws_xpixel / ws.ws_col;
+	*ppl = ws.ws_ypixel / ws.ws_row;
+	return 1;
+    }
 #endif
 
-static char XTERM_TITLE[] = "\033]0;w3m: %s\007";
-static char SCREEN_TITLE[] = "\033k%s\033\134";
-#ifdef __CYGWIN__
-static char CYGWIN_TITLE[] = "w3m: %s";
-#endif
+    fputs("\x1b[14t\x1b[18t",ttyf); flush_tty();
 
-/* *INDENT-OFF* */
-static struct w3m_term_info {
-    char *term;
-    char *title_str;
-#ifdef USE_MOUSE
-    int mouse_flag;
-#endif
-} w3m_term_info_list[] = {
-    {W3M_TERM_INFO("xterm", XTERM_TITLE, (NEED_XTERM_ON|NEED_XTERM_OFF))},
-    {W3M_TERM_INFO("kterm", XTERM_TITLE, (NEED_XTERM_ON|NEED_XTERM_OFF))},
-    {W3M_TERM_INFO("rxvt", XTERM_TITLE, (NEED_XTERM_ON|NEED_XTERM_OFF))},
-    {W3M_TERM_INFO("Eterm", XTERM_TITLE, (NEED_XTERM_ON|NEED_XTERM_OFF))},
-    {W3M_TERM_INFO("mlterm", XTERM_TITLE, (NEED_XTERM_ON|NEED_XTERM_OFF))},
-    {W3M_TERM_INFO("screen", SCREEN_TITLE, 0)},
-#ifdef __CYGWIN__
-    {W3M_TERM_INFO("cygwin", CYGWIN_TITLE, (NEED_CYGWIN_ON|NEED_CYGWIN_OFF))},
-#endif
-    {W3M_TERM_INFO(NULL, NULL, 0)}
-};
-#undef W3M_TERM_INFO
-/* *INDENT-ON * */
+    p = buf;
+    left = sizeof(buf) - 1;
+    for (i = 0; i < 10; i++) {
+	tval.tv_usec = 200000;	/* 0.2 sec * 10 */
+	tval.tv_sec = 0;
+	FD_ZERO(&rfd);
+	FD_SET(tty,&rfd);
+	if (select(tty+1,&rfd,NULL,NULL,&tval) <= 0 || ! FD_ISSET(tty,&rfd))
+	    continue;
+
+	if ((len = read(tty,p,left)) <= 0)
+	    continue;
+	p[len] = '\0';
+
+	if (sscanf(buf,"\x1b[4;%d;%dt\x1b[8;%d;%dt",&hp,&wp,&hc,&wc) == 4) {
+	    if (wp > 0 && wc > 0 && hp > 0 && hc > 0) {
+		*ppc = wp / wc;
+		*ppl = hp / hc;
+		return 1;
+	    }
+	    else {
+		return 0;
+	    }
+	}
+	p += len;
+	left -= len;
+    }
+
+    return 0;
+}
+#endif				/* USE_IMAGE */
+
+static const char *title_str_common = "\033]0;w3m: %s\007",
+    *title_str_screen = "\033k%s\033\134";
 
 int
 set_tty(void)
@@ -522,28 +851,17 @@ set_tty(void)
     check_cygwin_console();
 #endif
     TerminalGet(tty, &d_ioval);
-    if (displayTitleTerm != NULL) {
-	struct w3m_term_info *p;
-	for (p = w3m_term_info_list; p->term != NULL; p++) {
-	    if (!strncmp(displayTitleTerm, p->term, strlen(p->term))) {
-		title_str = p->title_str;
-		break;
-	    }
+    if (displayTitleTerm) {
+	if (!strncmp(displayTitleTerm, "screen", 6)) {
+	    title_str = title_str_screen;
+	} else {
+	    title_str = title_str_common;
 	}
     }
-#ifdef USE_MOUSE
-    {
-	char *term = getenv("TERM");
-	if (term != NULL) {
-	    struct w3m_term_info *p;
-	    for (p = w3m_term_info_list; p->term != NULL; p++) {
-		if (!strncmp(term, p->term, strlen(p->term))) {
-			is_xterm = p->mouse_flag;
-			break;
-		    }
-		}
+#ifdef USE_SYSMOUSE
+	if (!strncmp(getenv("TERM"), "screen", 6)){
+	    is_xterm = 0;
 	}
-    }
 #endif
     return 0;
 }
@@ -563,7 +881,7 @@ ttymode_set(int mode, int imode)
     while (TerminalSet(tty, &ioval) == -1) {
 	if (errno == EINTR || errno == EAGAIN)
 	    continue;
-	printf("Error occured while set %x: errno=%d\n", mode, errno);
+	printf("Error occurred while set %x: errno=%d\n", mode, errno);
 	reset_error_exit(SIGNAL_ARGLIST);
     }
 #endif
@@ -584,7 +902,7 @@ ttymode_reset(int mode, int imode)
     while (TerminalSet(tty, &ioval) == -1) {
 	if (errno == EINTR || errno == EAGAIN)
 	    continue;
-	printf("Error occured while reset %x: errno=%d\n", mode, errno);
+	printf("Error occurred while reset %x: errno=%d\n", mode, errno);
 	reset_error_exit(SIGNAL_ARGLIST);
     }
 #endif /* __MINGW32_VERSION */
@@ -601,7 +919,7 @@ set_cc(int spec, int val)
     while (TerminalSet(tty, &ioval) == -1) {
 	if (errno == EINTR || errno == EAGAIN)
 	    continue;
-	printf("Error occured: errno=%d\n", errno);
+	printf("Error occurred: errno=%d\n", errno);
 	reset_error_exit(SIGNAL_ARGLIST);
     }
 }
@@ -634,10 +952,11 @@ reset_tty(void)
     writestr(T_se);		/* reset terminal */
     flush_tty();
     TerminalSet(tty, &d_ioval);
-    close_tty();
+    if (tty != 2)
+        close_tty();
 }
 
-static MySignalHandler
+static void
 reset_exit_with_value(SIGNAL_ARG, int rval)
 {
 #ifdef USE_MOUSE
@@ -646,28 +965,26 @@ reset_exit_with_value(SIGNAL_ARG, int rval)
 #endif				/* USE_MOUSE */
     reset_tty();
     w3m_exit(rval);
-    SIGNAL_RETURN;
 }
 
-MySignalHandler
+void
 reset_error_exit(SIGNAL_ARG)
 {
   reset_exit_with_value(SIGNAL_ARGLIST, 1);
 }
 
-MySignalHandler
+void
 reset_exit(SIGNAL_ARG)
 {
   reset_exit_with_value(SIGNAL_ARGLIST, 0);
 }
 
-MySignalHandler
+void
 error_dump(SIGNAL_ARG)
 {
     mySignal(SIGIOT, SIG_DFL);
     reset_tty();
     abort();
-    SIGNAL_RETURN;
 }
 
 void
@@ -717,8 +1034,7 @@ getTCstr(void)
     char *pt = funcstr;
     int r;
 
-    ent = getenv("TERM") ? getenv("TERM") : DEFAULT_TERM;
-    if (ent == NULL) {
+    if (!(ent = getenv("TERM"))) {
 	fprintf(stderr, "TERM is not set\n");
 	reset_error_exit(SIGNAL_ARGLIST);
     }
@@ -768,12 +1084,6 @@ getTCstr(void)
     GETSTR(T_op, "op");		/* set default color pair to its original value */
 #if defined( CYGWIN ) && CYGWIN < 1
     /* for TERM=pcansi on MS-DOS prompt. */
-#if 0
-    T_eA = "";
-    T_as = "\033[12m";
-    T_ae = "\033[10m";
-    T_ac = "l\001k\002m\003j\004x\005q\006n\020a\024v\025w\026u\027t\031";
-#endif
     T_eA = "";
     T_as = "";
     T_ae = "";
@@ -822,10 +1132,10 @@ setlinescols(void)
 	LINES = tgetnum("li");	/* number of line */
     if (COLS <= 0)
 	COLS = tgetnum("co");	/* number of column */
-    if (COLS > MAX_COLUMN)
-	COLS = MAX_COLUMN;
-    if (LINES > MAX_LINE)
-	LINES = MAX_LINE;
+    if (MaxCols && COLS > MaxCols)
+	COLS = MaxCols;
+    if (opt_cols && COLS > opt_cols)
+	COLS = opt_cols;
 #if defined(__CYGWIN__)
     LASTLINE = LINES - (isWinConsole == TERM_CYGWIN_RESERVE_IME ? 2 : 1);
 #endif				/* defined(__CYGWIN__) */
@@ -847,7 +1157,7 @@ setupscreen(void)
 	for (i = 0; i < max_LINES; i++) {
 #ifdef USE_M17N
 	    ScreenElem[i].lineimage = New_N(char *, max_COLS);
-	    bzero((void *)ScreenElem[i].lineimage, max_COLS * sizeof(char *));
+	    bzero(ScreenElem[i].lineimage, max_COLS * sizeof(char *));
 #else
 	    ScreenElem[i].lineimage = New_N(char, max_COLS);
 #endif
@@ -883,7 +1193,7 @@ initscr(void)
 }
 
 static int
-write1(char c)
+write1(const char c)
 {
     putc(c, ttyf);
 #ifdef SCREEN_DEBUG
@@ -909,13 +1219,13 @@ move(int line, int column)
 
 static int
 #ifdef USE_M17N
-need_redraw(char *c1, l_prop pr1, char *c2, l_prop pr2)
+need_redraw(const char *c1, l_prop pr1, const char *c2, l_prop pr2)
 {
     if (!c1 || !c2 || strcmp(c1, c2))
 	return 1;
     if (*c1 == ' ')
 #else
-need_redraw(char c1, l_prop pr1, char c2, l_prop pr2)
+need_redraw(const char c1, l_prop pr1, const char c2, l_prop pr2)
 {
     if (c1 != c2)
 	return 1;
@@ -939,26 +1249,25 @@ need_redraw(char c1, l_prop pr1, char c2, l_prop pr2)
 
 #ifdef USE_M17N
 void
-addch(char c)
+addch(const char c)
 {
     addmch(&c, 1);
 }
 
 void
-addmch(char *pc, size_t len)
+addmch(const char *pc, size_t len)
 #else
 void
-addch(char pc)
+addch(const char pc)
 #endif
 {
     l_prop *pr;
     int dest, i;
-    short *dirty;
 #ifdef USE_M17N
     static Str tmp = NULL;
     char **p;
     char c = *pc;
-    int width = wtf_width((wc_uchar *) pc);
+    int width = wtf_width((const wc_uchar *) pc);
 
     if (tmp == NULL)
 	tmp = Strnew();
@@ -975,7 +1284,6 @@ addch(char pc)
 	return;
     p = ScreenImage[CurLine]->lineimage;
     pr = ScreenImage[CurLine]->lineprop;
-    dirty = &ScreenImage[CurLine]->isdirty;
 
 #ifndef USE_M17N
     /* Eliminate unprintables according to * iso-8859-*.
@@ -1229,8 +1537,12 @@ setfcolor(int color)
 static char *
 color_seq(int colmode)
 {
+    int n = 30;
     static char seqbuf[32];
-    sprintf(seqbuf, "\033[%dm", ((colmode >> 8) & 7) + 30);
+#ifdef USE_COLOR
+    n = highIntensityColors ? 90 : 30;
+#endif
+    sprintf(seqbuf, "\033[%dm", ((colmode >> 8) & 7) + n);
     return seqbuf;
 }
 
@@ -1570,24 +1882,6 @@ rscroll(int n)
 }
 #endif
 
-#if 0
-void
-need_clrtoeol(void)
-{
-    /* Clear to the end of line as the need arises */
-    l_prop *lprop = ScreenImage[CurLine]->lineprop;
-
-    if (lprop[CurColumn] & S_EOL)
-	return;
-
-    if (!(ScreenImage[CurLine]->isdirty & (L_NEED_CE | L_CLRTOEOL)) ||
-	ScreenImage[CurLine]->eol > CurColumn)
-	ScreenImage[CurLine]->eol = CurColumn;
-
-    ScreenImage[CurLine]->isdirty |= L_NEED_CE;
-}
-#endif				/* 0 */
-
 /* XXX: conflicts with curses's clrtoeol(3) ? */
 void
 clrtoeol(void)
@@ -1610,7 +1904,7 @@ clrtoeol(void)
 }
 
 #ifdef USE_BG_COLOR
-void
+static void
 clrtoeol_with_bcolor(void)
 {
     int i, cli, cco;
@@ -1644,8 +1938,8 @@ clrtoeolx(void)
 }
 #endif				/* not USE_BG_COLOR */
 
-void
-clrtobot_eol(void (*clrtoeol) ())
+static void
+clrtobot_eol(void (*clrtoeol) (void))
 {
     int l, c;
 
@@ -1660,37 +1954,20 @@ clrtobot_eol(void (*clrtoeol) ())
     CurColumn = c;
 }
 
-void
-clrtobot(void)
-{
-    clrtobot_eol(clrtoeol);
-}
-
 void
 clrtobotx(void)
 {
     clrtobot_eol(clrtoeolx);
 }
 
-#if 0
 void
-no_clrtoeol(void)
-{
-    int i;
-    l_prop *lprop = ScreenImage[CurLine]->lineprop;
-
-    ScreenImage[CurLine]->isdirty &= ~L_CLRTOEOL;
-}
-#endif				/* 0 */
-
-void
-addstr(char *s)
+addstr(const char *s)
 {
 #ifdef USE_M17N
     int len;
 
     while (*s != '\0') {
-	len = wtf_len((wc_uchar *) s);
+	len = wtf_len((const wc_uchar *) s);
 	addmch(s, len);
 	s += len;
     }
@@ -1701,17 +1978,17 @@ addstr(char *s)
 }
 
 void
-addnstr(char *s, int n)
+addnstr(const char *s, int n)
 {
     int i;
 #ifdef USE_M17N
     int len, width;
 
     for (i = 0; *s != '\0';) {
-	width = wtf_width((wc_uchar *) s);
+	width = wtf_width((const wc_uchar *) s);
 	if (i + width > n)
 	    break;
-	len = wtf_len((wc_uchar *) s);
+	len = wtf_len((const wc_uchar *) s);
 	addmch(s, len);
 	s += len;
 	i += width;
@@ -1723,17 +2000,17 @@ addnstr(char *s, int n)
 }
 
 void
-addnstr_sup(char *s, int n)
+addnstr_sup(const char *s, int n)
 {
     int i;
 #ifdef USE_M17N
     int len, width;
 
     for (i = 0; *s != '\0';) {
-	width = wtf_width((wc_uchar *) s);
+	width = wtf_width((const wc_uchar *) s);
 	if (i + width > n)
 	    break;
-	len = wtf_len((wc_uchar *) s);
+	len = wtf_len((const wc_uchar *) s);
 	addmch(s, len);
 	s += len;
 	i += width;
@@ -1764,29 +2041,6 @@ crmode(void)
 }
 #endif				/* HAVE_SGTTY_H */
 
-void
-nocrmode(void)
-#ifndef HAVE_SGTTY_H
-{
-    ttymode_set(ICANON, 0);
-#ifdef HAVE_TERMIOS_H
-    set_cc(VMIN, 4);
-#else				/* not HAVE_TERMIOS_H */
-    set_cc(VEOF, 4);
-#endif				/* not HAVE_TERMIOS_H */
-}
-#else				/* HAVE_SGTTY_H */
-{
-    ttymode_reset(CBREAK, 0);
-}
-#endif				/* HAVE_SGTTY_H */
-
-void
-term_echo(void)
-{
-    ttymode_set(ECHO, 0);
-}
-
 void
 term_noecho(void)
 {
@@ -1847,11 +2101,17 @@ term_cbreak(void)
 }
 
 void
-term_title(char *s)
+term_title(const char *s)
 {
     if (!fmInitialized)
         return;
     if (title_str != NULL) {
+/*
+ * TODO(chimera lover):
+ * broken, should rm once SUPPORT_WIN9X_CONSOLE_MBCS and
+ * TERM=cygwin special handle announced deprecation and
+ * no one complains
+ */
 #ifdef __CYGWIN__
 	if (isLocalConsole && title_str == CYGWIN_TITLE) {
 	    Str buff;
@@ -1907,18 +2167,15 @@ wgetch(void *p)
 }
 
 int
-do_getch()
+do_getch(void)
 {
-    if (is_xterm || !gpm_handler)
 	return getch();
-    else
-	return Gpm_Getch();
 }
 #endif				/* USE_GPM */
 
 #ifdef USE_SYSMOUSE
 int
-sysm_getch()
+sysm_getch(void)
 {
     fd_set rfd;
     int key, x, y;
@@ -1938,7 +2195,7 @@ sysm_getch()
 }
 
 int
-do_getch()
+do_getch(void)
 {
     if (is_xterm || !sysm_handler)
 	return getch();
@@ -1946,7 +2203,7 @@ do_getch()
 	return sysm_getch();
 }
 
-MySignalHandler
+void
 sysmouse(SIGNAL_ARG)
 {
     struct mouse_info mi;
@@ -1973,7 +2230,7 @@ bell(void)
     write1(7);
 }
 
-void
+static void
 skip_escseq(void)
 {
     int c;
@@ -1982,10 +2239,14 @@ skip_escseq(void)
     if (c == '[' || c == 'O') {
 	c = getch();
 #ifdef USE_MOUSE
-	if (is_xterm && c == 'M') {
+	if (c == 'M') {
 	    getch();
 	    getch();
 	    getch();
+	} else if (c == '<') {
+	    c = getch();
+	    while (IS_DIGIT(c) || c == ';')
+		c = getch();
 	}
 	else
 #endif
@@ -2019,24 +2280,20 @@ sleep_till_anykey(int sec, int purge)
     }
     er = TerminalSet(tty, &ioval);
     if (er == -1) {
-	printf("Error occured: errno=%d\n", errno);
+	printf("Error occurred: errno=%d\n", errno);
 	reset_error_exit(SIGNAL_ARGLIST);
     }
     return ret;
 }
 
 #ifdef USE_MOUSE
-
-#define XTERM_ON   {fputs("\033[?1001s\033[?1000h",ttyf); flush_tty();}
-#define XTERM_OFF  {fputs("\033[?1000l\033[?1001r",ttyf); flush_tty();}
-#define CYGWIN_ON  {fputs("\033[?1000h",ttyf); flush_tty();}
-#define CYGWIN_OFF {fputs("\033[?1000l",ttyf); flush_tty();}
-
+#define MOUSE_ON  {fputs("\033[?1001s\033[?1000h\033[?1006h",ttyf); flush_tty();}
+#define MOUSE_OFF {fputs("\033[?1006l\033[?1000l\033[?1001r",ttyf); flush_tty();}
 #ifdef USE_GPM
 /* Linux console with GPM support */
 
 void
-mouse_init()
+mouse_init(void)
 {
     Gpm_Connect conn;
     extern int gpm_process_mouse(Gpm_Event *, void *);
@@ -2059,36 +2316,28 @@ mouse_init()
 	 * xterm is being left in the mode where the mouse clicks are
 	 * passed through to the application.
 	 */
-	Gpm_Close();
-	is_xterm = (NEED_XTERM_ON | NEED_XTERM_OFF);
     }
     else if (r >= 0) {
 	gpm_handler = gpm_process_mouse;
-	is_xterm = 0;
-    }
-    if (is_xterm) {
-	XTERM_ON;
     }
+    Gpm_Close();
+    MOUSE_ON;
     mouseActive = 1;
 }
 
 void
-mouse_end()
+mouse_end(void)
 {
     if (mouseActive == 0)
 	return;
-    if (is_xterm) {
-	XTERM_OFF;
-    }
-    else
-	Gpm_Close();
+    MOUSE_OFF;
     mouseActive = 0;
 }
 
 #elif	defined(USE_SYSMOUSE)
 /* *BSD console with sysmouse support */
 void
-mouse_init()
+mouse_init(void)
 {
     mouse_info_t mi;
     extern int sysm_process_mouse();
@@ -2096,7 +2345,7 @@ mouse_init()
     if (mouseActive)
 	return;
     if (is_xterm) {
-	XTERM_ON;
+	MOUSE_ON;
     }
     else {
 #if defined(FBIO_MODEINFO) || defined(CONS_MODEINFO)	/* FreeBSD > 2.x */
@@ -2129,12 +2378,12 @@ mouse_init()
 }
 
 void
-mouse_end()
+mouse_end(void)
 {
     if (mouseActive == 0)
 	return;
     if (is_xterm) {
-	XTERM_OFF;
+	MOUSE_OFF;
     }
     else {
 	mouse_info_t mi;
@@ -2150,34 +2399,20 @@ mouse_end()
 /* not GPM nor SYSMOUSE, but use mouse with xterm */
 
 void
-mouse_init()
+mouse_init(void)
 {
     if (mouseActive)
 	return;
-    if (is_xterm & NEED_XTERM_ON) {
-	XTERM_ON;
-    }
-#ifdef __CYGWIN__
-    else if (is_xterm & NEED_CYGWIN_ON) {
-	CYGWIN_ON;
-    }
-#endif
+    MOUSE_ON;
     mouseActive = 1;
 }
 
 void
-mouse_end()
+mouse_end(void)
 {
     if (mouseActive == 0)
 	return;
-    if (is_xterm & NEED_XTERM_OFF) {
-	XTERM_OFF;
-    }
-#ifdef __CYGWIN__
-    else if (is_xterm & NEED_CYGWIN_OFF) {
-	CYGWIN_OFF;
-    }
-#endif
+    MOUSE_OFF;
     mouseActive = 0;
 }
 
@@ -2185,23 +2420,23 @@ mouse_end()
 
 
 void
-mouse_active()
+mouse_active(void)
 {
     if (!mouseActive)
 	mouse_init();
 }
 
 void
-mouse_inactive()
+mouse_inactive(void)
 {
-    if (mouseActive && is_xterm)
+    if (mouseActive)
 	mouse_end();
 }
 
 #endif				/* USE_MOUSE */
 
 void
-flush_tty()
+flush_tty(void)
 {
     if (ttyf)
 	fflush(ttyf);
@@ -2209,7 +2444,7 @@ flush_tty()
 
 #ifdef USE_IMAGE
 void
-touch_cursor()
+touch_cursor(void)
 {
 #ifdef USE_M17N
     int i;
